// XGS mock-up GPU emulator
// Mic, 2004

/*
	VRAM layout is as follows:

	Bitmapped mode:
		
		Start	Size		Contents
		---------------------------------
		0x00000	0x12C00		Bitmap data	
		0x12C00	0x00200		Palette (256 colors, 2 bytes each (RGB565)
		0x12E00	0x00200		Sprite table
		0x20000	0x10000		Tile bank #2

	Tiled mode:

		Start	Size		Contents
		----------------------------
		0x00000	0x10000		Tile bank #1. 256 16x16 tiles
		0x10000	0x01000		Name table. 64*32 tiles, 2 bytes/entry
		0x12C00	0x00200		Palette
		0x12E00	0x00200		Sprite table. 128 sprites, 4 bytes/entry
		0x20000	0x10000		Tile bank #2

	Name table format:	ppppuvht tttttttt
	-----------------
	t = tile number (0-511)
	h = flip horizontally if set
	v = flip vertically if set
	p = palette (ignored in 8bpp mode)
	u = unused

	Sprite table format:	ppppvhtt tttttttx xxxxxxxx yyyyyyyy
	-------------------
	x = x coordinate (9 bits)
	y = y coordinate (8 bits)
	t = tile number (9 bits)
	h = flip horizontally
	v = flip vertically
	p = palette (ignored in 8bpp mode)

	GPU registers:
	--------------
	reg0	vhdbtccu	v = generate vblank interrupts
						h = generate hblank interrupts
						d = display enabled
						b = bitmapped mode if set
						t = 1=tiles are 16x16, 0=tiles are 8x8
						c = color format (10=8bpp, 01=4bpp, 00=2bpp)
						u = unused

	reg1	vuuuuuuu	v = in vblank

	reg2	llllllll	l = current scanline

	reg4	hhhhhhhh	h = horizontal scroll

	reg5	vvvvvvvv	v = vertical scroll
*/


#include "zombie.h"
#include <commctrl.h>
#include <string.h>
#include "resrc1.h"

#define JUMP	0
#define VECTOR	1

#define NAME_TABLE_OFFSET	0x10000
#define PALETTE_OFFSET		0x12C00
#define SPRITE_TABLE_OFFSET	0x12E00


BOOL CALLBACK DlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam);
void draw_scanline();
void finish_screen();
void reset_screen();
int __stdcall inpb_00();
int __stdcall inpb_01();
int __stdcall inpb_02();
int __stdcall inpb_03();
int __stdcall inpb_04();
int __stdcall inpb_05();
int __stdcall inpb_06();
int __stdcall inpb_07();
void __stdcall outpb_00(int);
void __stdcall outpb_01(int);
void __stdcall outpb_02(int);
void __stdcall outpb_03(int);
void __stdcall outpb_04(int);
void __stdcall outpb_05(int);
void __stdcall outpb_06(int);
void __stdcall outpb_07(int);
void __stdcall outpb_10(int);
void __stdcall outpb_11(int);



HANDLE hout;
HINSTANCE hInst;
HWND hwnd;
HDC dc;
RECT rect;
char *VRAM,*vscreen,*patternTbl;
char dataLatch;
short int *colorLUT;
int width,scanline,visibleLines,totalLines,nameTblHeight;
int hbiAction,hbiAddress,vbiAction,vbiAddress;
int yscroll;
int vramPtr,addressLatch;
int spriteCnt,spriteLimit;
int htile,patLine,patLineInv,tileShift;
int addressShift;
int dummy;
DWORD lastFrameTime;
bool hblankEnabled,vblankEnabled;
char gpuName[] = "XGS GPU";

cpu_irq_callback cpu_irq;
cpu_set_io_handler_callback cpu_set_io_handler;
mmu_read_word_callback mmu_read_word;
zombie_callback zombie;

__int64 fifteen = 15;
int plus1=1,minus1=-1;
LARGE_INTEGER perfFreq,perfCnt1,perfCnt2;

palette_dib palDib;


// Used to copy the screen buffer to the window
struct bi
{
	BITMAPINFOHEADER bih;
	RGBQUAD mask[3];
}bi;


// GPU registers
struct gpu
{
	unsigned char reg0,reg1,reg2,reg3,reg4,reg5,reg6,reg7;
}gpu;





bool __stdcall DllMain(HINSTANCE hDLL,DWORD reason,LPVOID unused)
{
	if (reason==DLL_PROCESS_ATTACH)
		hInst = hDLL;

	return true;
}




bool __stdcall gpu_init(zombie_callback commlink)
{
	zombie = commlink;

	scanline = 0;
	width = 320;
	visibleLines = 240;
	totalLines = 262;
	nameTblHeight = 240;
	spriteLimit = 255;

	hblankEnabled = true;
	vblankEnabled = true;

	hbiAction = vbiAction = JUMP;
	hbiAddress = vbiAddress = 0x38;

	VRAM = (char*)LocalAlloc(LPTR,0x40000);			// 256 kB of VRAM
	vscreen = (char*)LocalAlloc(LPTR,320*240*2);
	colorLUT = (short*)&VRAM[0x12C00];
	patternTbl = VRAM;


	// Ask for the handle of the parent window
	hwnd = (HWND)zombie(ASK,ZOMBIE_WINDOW_HANDLE,0,0);
	dc = GetDC(hwnd);

	// Get a handle to STDOUT (for WriteConsole)
	hout = GetStdHandle(STD_OUTPUT_HANDLE);

	return true;
}



bool __stdcall gpu_reset()
{
	*(__int64*)&gpu = 0;

	// Set I/O handlers
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x00|(INPUT<<8)),inpb_00);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x01|(INPUT<<8)),inpb_01);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x02|(INPUT<<8)),inpb_02);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x03|(INPUT<<8)),inpb_03);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x04|(INPUT<<8)),inpb_04);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x05|(INPUT<<8)),inpb_05);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x06|(INPUT<<8)),inpb_06);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x07|(INPUT<<8)),inpb_07);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x00|(OUTPUT<<8)),outpb_00);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x01|(OUTPUT<<8)),outpb_01);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x02|(OUTPUT<<8)),outpb_02);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x02|(OUTPUT<<8)),outpb_03);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x04|(OUTPUT<<8)),outpb_04);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x05|(OUTPUT<<8)),outpb_05);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x06|(OUTPUT<<8)),outpb_06);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x07|(OUTPUT<<8)),outpb_07);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x10|(OUTPUT<<8)),outpb_10);
	zombie(TELL,CPU_IO_HANDLER,(void*)(0x11|(OUTPUT<<8)),outpb_11);

	// Map GPU ports to CPU memory space
	zombie(TELL,MMU_MAP_PORT,(void*)0x00,(void*)0xFF00);
	zombie(TELL,MMU_MAP_PORT,(void*)0x01,(void*)0xFF01);
	zombie(TELL,MMU_MAP_PORT,(void*)0x02,(void*)0xFF02);
	zombie(TELL,MMU_MAP_PORT,(void*)0x03,(void*)0xFF03);
	zombie(TELL,MMU_MAP_PORT,(void*)0x04,(void*)0xFF04);
	zombie(TELL,MMU_MAP_PORT,(void*)0x05,(void*)0xFF05);
	zombie(TELL,MMU_MAP_PORT,(void*)0x06,(void*)0xFF06);
	zombie(TELL,MMU_MAP_PORT,(void*)0x07,(void*)0xFF07);
	zombie(TELL,MMU_MAP_PORT,(void*)0x10,(void*)0xFF10);
	zombie(TELL,MMU_MAP_PORT,(void*)0x11,(void*)0xFF11);

	// Ask for the CPU IRQ callback
	cpu_irq = (cpu_irq_callback)zombie(ASK,CPU_IRQ_CALLBACK,0,0);

	memset(VRAM,0,0x40000);
	
	addressShift = 0;
	//lastFrameTime = GetTickCount();
	QueryPerformanceFrequency(&perfFreq);
	QueryPerformanceCounter(&perfCnt1);
	return true;
}


void __stdcall gpu_close()
{
	LocalFree(vscreen);
	LocalFree(VRAM);
}


int __stdcall gpu_config(HWND hwnd)
{
	return 0;
	/*return DialogBoxParam(hInst,
           MAKEINTRESOURCE(IDD_GPU_CONFIG), hwnd,
           DlgProc, NULL);*/
}


int __stdcall gpu_get(int what)
{
	switch (what)
	{
		case GPU_CURRENT_SCANLINE:
			return scanline;
		case GPU_SCREEN_WIDTH:
			return width;
		case GPU_SCREEN_HEIGHT:
			return visibleLines;
		case GPU_VBLANK_LINES:
			return totalLines-visibleLines;
		case GPU_HBLANK_PIXELS:
			return 32;
		case GPU_NUMBER_COLORS:
			return 256;
		case GPU_PALETTE:
			palDib.data = &VRAM[PALETTE_OFFSET];
			memset(&palDib.bih,0,sizeof(BITMAPINFOHEADER));
 			palDib.bih.biSize = sizeof(BITMAPINFOHEADER);
 			palDib.bih.biWidth = 32;
 			palDib.bih.biHeight = -8;
 			palDib.bih.biPlanes = 1;
 			palDib.bih.biBitCount = 16;
  			palDib.bih.biCompression = BI_BITFIELDS;

			*((int*)&palDib.colors[0]) = 0xF800;	// RRRRR000 00000000
			*((int*)&palDib.colors[1]) = 0x07E0;	// 00000GGG GGG00000
			*((int*)&palDib.colors[2]) = 0x001F;	// 00000000 000BBBBB

			return (int)&palDib;

		case GPU_MEMORY:
			return (int)VRAM;
		case PLUGIN_NAME:
			return (int)&gpuName;
		default:
			break;
	}

	return 0;
}


void __stdcall gpu_set(int what,int param1,int param2)
{
	switch (what)
	{
		case CPU_IRQ_CALLBACK:
			cpu_irq = (cpu_irq_callback)param1;
			break;
		case MMU_READ_WORD_CALLBACK:
			mmu_read_word = (mmu_read_word_callback)param1;
			break;
		case GPU_VBLANK_ACTION:
				vbiAction = param1;
				vbiAddress = param2;
				break;
		case GPU_HBLANK_ACTION:
				hbiAction = param1;
				hbiAddress = param2;
				break;
		case ZOMBIE_WINDOW_HANDLE:
			hwnd = (HWND)param1;
			break;
		default:
			break;
	}
}


void __stdcall gpu_hblank()
{
	if (hblankEnabled && ((gpu.reg0 & 0x40) != 0) && scanline<visibleLines)
	{
		if (hbiAction == JUMP)
			cpu_irq(hbiAddress);
		else
			cpu_irq(mmu_read_word(hbiAddress));
	}
}



void __stdcall gpu_advance_line()
{
	static char buffer[80];

	gpu.reg2 = scanline;
	scanline++;

	if (scanline<=visibleLines) 
		draw_scanline();
	else if (scanline==visibleLines+1)
	{
		gpu.reg1 |= 0x80;
		if (vblankEnabled && ((gpu.reg0 & 0x80) != 0))
			if (vbiAction == JUMP)
				cpu_irq(vbiAddress);
			else
				cpu_irq(mmu_read_word(vbiAddress));
	}
	else if (scanline==totalLines)
	{
		finish_screen();
		reset_screen();
		gpu.reg2 = 0;
		scanline = 0;
		gpu.reg1 &= 0x7F;
	}
}



// IO handlers

int __stdcall inpb_00()
{
	return gpu.reg0;
}


int __stdcall inpb_01()
{
	char r;

	r = gpu.reg1;

	// Reading reg1 clears vblank flag
	gpu.reg1 &= 0x7F;

	return (int)r;
}

int __stdcall inpb_02()
{
	return gpu.reg2;
}

int __stdcall inpb_03()
{
	return gpu.reg3;
}

int __stdcall inpb_04()
{
	return gpu.reg4;
}

int __stdcall inpb_05()
{
	return gpu.reg5;
}

int __stdcall inpb_06()
{
	return gpu.reg6;
}

int __stdcall inpb_07()
{
	return gpu.reg7;
}

int __stdcall inpb_10()
{
	return 0;
}

int __stdcall inpb_11()
{
	// Return the last byte written
	return (int)dataLatch;
}


void __stdcall outpb_00(int byte)
{
	static char buffer[64];

	gpu.reg0 = (char)byte;
}
void __stdcall outpb_01(int byte)
{
	gpu.reg1 = (char)byte;
}
void __stdcall outpb_02(int byte)
{
	gpu.reg2 = (char)byte;
}
void __stdcall outpb_03(int byte)
{
	gpu.reg3 = (char)byte;
}
void __stdcall outpb_04(int byte)
{
	gpu.reg4 = (char)byte;
}
void __stdcall outpb_05(int byte)
{
	gpu.reg5 = (char)byte;
}
void __stdcall outpb_06(int byte)
{
	gpu.reg6 = (char)byte;
}
void __stdcall outpb_07(int byte)
{
	gpu.reg7 = (char)byte;
}


// GPU address port
void __stdcall outpb_10(int byte)
{
	addressLatch = (byte<<addressShift)|addressLatch;

	addressShift += 8;
	// Only set the VRAM pointer when we have a full 32-bit address
	if (addressShift == 32)
	{
		vramPtr = addressLatch;
		addressShift = 0;
		addressLatch = 0;
	}
}


// GPU data port
void __stdcall outpb_11(int byte)
{
	static char buffer[80];

	/*if (vramPtr>=SPRITE_TABLE_OFFSET)
	{
		wsprintf(buffer,"Writing %02x to %04x\n",byte,vramPtr);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long *)&dummy,NULL);
	}*/

	dataLatch = (char)byte;
	VRAM[vramPtr++] = (char)byte;
	vramPtr &= 0x3FFFF;
}



void draw_scanline()
{
	int tilebias = 0;

	// Return if the display is disabled
	if ((gpu.reg0 & 0x20) == 0)
		return;

	tileShift = 6 + ((gpu.reg0 & 6)>>1);

	// Is bitmapped mode enabled?
	if ((gpu.reg0 & 0x10) != 0)
	{

		__asm
		{
			pusha
			mov eax,[scanline]
			dec eax
			mov edi,[vscreen]
			shl eax,7
			mov esi,[VRAM]
			lea eax,[eax+eax*4]
			add edi,eax				// edi = vscreen+scanline*640
			shr eax,1
			add esi,eax				// esi = VRAM+scanline*320

			mov ebx,[VRAM]
			add ebx,PALETTE_OFFSET
			add esi,320				// The add is done to allow negative indexing
			add edi,640				// ..dito
			mov ecx,-320
			__bm_pixel_loop:
				xor eax,eax
				mov al,[esi+ecx]	// Read one byte from VRAM
				mov dx,[ebx+eax*2]	// Look it up in the color table
				mov [edi+ecx*2],dx	// Store it in the virtual screen
				inc ecx				// Next pixel..
			jnz __bm_pixel_loop
		}

	} else
	{

		// Tiled mode
		__asm
		{
			pusha
			mov eax,[scanline]
			cld
			dec eax
			shl eax,7
			lea eax,[eax+eax*4]
			add eax,[vscreen]
			mov edi,eax 			// edi = scanline*320*2 + vscreen

			movd mm0,edi	 		// save edi
			mov eax,[VRAM]
			add eax,PALETTE_OFFSET
			movzx ebx,word ptr [eax] // get background color (color 0)
			mov eax,ebx
			shl ebx,16
			or eax,ebx
			mov ecx,160 			// clear the scanline
			rep stosd
			movd edi,mm0			// restore edi

			movzx ebx,byte ptr [gpu.reg4]
			mov edx,ebx
			and ebx,15 				// ebx = hscroll%15
			shr edx,4 				// edx = # of tiles to scroll

			mov eax,[scanline]
			dec eax
			add eax,[yscroll]
			cmp eax,[nameTblHeight]
			jl __line_ok
			sub eax,[nameTblHeight]
			__line_ok:

			movd mm0,eax
			pand mm0,qword ptr [fifteen]
			psllw mm0,4
			movd [patLine],mm0
			movd [patLineInv],mm0
			movd mm1,[patternTbl]
			xor word ptr [patLineInv],0xF0
			paddd mm0,mm1			// mm0 = ((line+yscroll)&15)*16 + patternTbl

			shr eax,4				// name table row
			shl eax,6				// 64 tiles per row
			add eax,edx
			add eax,eax
			add eax,NAME_TABLE_OFFSET
			add eax,[VRAM]
			mov esi,eax 			// ((line/16)*64 + hscroll)*2 + NAME_TABLE_OFFSET + VRAM

			mov dword ptr [htile],0

			__hloop:
			cmp dword ptr [htile],0
			jne __not_first_colm
			movzx eax,word ptr [esi]

			mov edx,eax
			and edx,0xF000
			shr edx,7
			add edx,PALETTE_OFFSET
			add edx,[VRAM]
			mov [colorLUT],edx

			mov edx,eax
			and eax,0x1FF
	
			mov ecx,[tileShift]	
			shl eax,cl 				// tile*2^bpp

			mov ecx,16
			sub ecx,ebx 			// first tile, length is 16-(hscroll%15)

			movd mm1,eax
			paddd mm1,mm0
			movd eax,mm1
			add eax,ebx 			// tile*2^bpp + (line&15)*16 + patternTbl + hscroll%15
			jmp __draw_pixels
			__first_colm_hidden:

			__not_first_colm:
			cmp dword ptr [htile],20
			je __last_colm
			mov ecx,16 				// "normal" tile consisting of 8 pixels
			jmp __l1

			__last_colm:
			mov ecx,ebx 			// last tile, length is hscroll%15
			__l1:
			movzx eax,word ptr [esi]
			mov edx,eax
			movd mm2,ecx			// Save ecx
			and eax,0x1FF
			mov ecx,[tileShift]
			shl eax,cl
			movd mm1,eax
			movd ecx,mm2			// restore ecx
			paddd mm1,mm0
			movd eax,mm1

			__draw_pixels:
			movd mm2,[plus1]
			test edx,0x200
			jz __no_hflip
			movd mm2,[minus1]
			lea eax,[eax+ecx-1]
			__no_hflip:
			test edx,0x400
			jz __no_vflip
			sub eax,[patLine]
			add eax,[patLineInv]
			__no_vflip:

			movd mm3,eax
			push ebx
			push esi
			mov esi,[colorLUT]

			cmp dword ptr [tileShift],8
			jne __maybe_4bpp

			__pixel_loop_8:
				and ecx,ecx
				jz __next_tile
				movd eax,mm3
				xor ebx,ebx
				mov bl,[eax]
				paddd mm3,mm2
				mov bx,[ebx*2+esi]
				mov [edi],bx
				dec ecx
				add edi,2
			jmp __pixel_loop_8

			__maybe_4bpp:
			cmp dword ptr [tileShift],7
			jne __maybe_2bpp

			__pixel_loop_4:
				and ecx,ecx
				jz __next_tile
				movd eax,mm3
				xor ebx,ebx
				mov bl,[eax]
				paddd mm3,mm2
				movd mm4,ebx
				and ebx,0x0F
				mov bx,[ebx*2+esi]
				mov [edi],bx
				movd ebx,mm4
				shr ebx,4
				mov bx,[ebx*2+esi]
				mov [edi+2],bx
				sub ecx,2
				add edi,4
			jmp __pixel_loop_4

			__maybe_2bpp:
			__next_tile:
			pop esi
			pop ebx
			add esi,2
			inc dword ptr [htile]
			cmp dword ptr [htile],21	// Quit when we've drawn 20 tiles (320/16)
			jne near __hloop
		}
	}

	// Sprite-drawing code
	__asm
	{
		__draw_sprites:
		dec dword ptr [scanline]
		mov eax,[scanline]
		shl eax,7
		lea eax,[eax+eax*4]
		add eax,[vscreen]
		mov edi,eax

		mov ecx,0
		mov dword ptr [spriteCnt],0		// # of sprites on this scanline
		mov edx,[VRAM]
		add edx,SPRITE_TABLE_OFFSET
		__sprite_loop:
		mov eax,[edx+ecx*4] 	
		mov esi,15

		mov ebx,eax
		and ebx,0xF0000000
		shr ebx,7
		add ebx,PALETTE_OFFSET
		add ebx,[VRAM]
		mov [colorLUT],ebx

		movd mm0,eax
		and eax,0xFF					// eax = sprite.y
		lea ebx,[eax+esi] 				// ebx = sprite.y+(height-1)
		cmp eax,dword ptr [scanline] 	// is the entire sprite below this scanline?
		jg __next_sprite
		cmp ebx,dword ptr [scanline] 	// is the entire sprite above this scanline?
		jl __next_sprite

		movd mm1,eax
		mov eax,[spriteLimit]
		inc dword ptr [spriteCnt]
		cmp dword ptr [spriteCnt],eax
		jg __sprites_done
		movd eax,mm1

		push edi 					// save edi
		mov ebx,[scanline]
		sub ebx,eax
		shl ebx,4 					// ebx = (line-sprite.y)*16
		movd eax,mm0
		mov esi,ebx
		add esi,[patternTbl]
		mov ebx,eax
		and eax,0x3FE0000
		shr eax,9					// tile#
		test eax,0x10000
		jz __not_high_bank
			add esi,0x20000
		__not_high_bank:
		and eax,0xFF00

		and ebx,0x1FF00
		shr ebx,8					// sprite.x

		add esi,eax 				// eax = tile*256 + (line-sprite.y)*16 + patternTbl
		lea edi,[edi+ebx*2]			// edi += sprite.x*2
		push ecx
		mov ebx,[colorLUT]
		mov ecx,16
		__sprite_hloop_8:
			xor eax,eax
			mov al,[esi]
			or eax,eax
			jz __sprite_pixel_invisible
			mov ax,[ebx+eax*2]
			mov [edi],ax
			__sprite_pixel_invisible:
			inc esi
			add edi,2
			dec ecx
		jnz __sprite_hloop_8
		pop ecx
		pop edi

		__next_sprite:
		inc ecx
		cmp ecx,64
		jne near __sprite_loop
		__sprites_done:
		popa
		inc dword ptr [scanline]
		emms
	}
}



void finish_screen()
{
	if ((gpu.reg0 & 0x20) == 0)
		return;
	
	QueryPerformanceCounter(&perfCnt2);
	while (((perfCnt2.QuadPart-perfCnt1.QuadPart)<<16)/perfFreq.QuadPart < 1092)
	{ QueryPerformanceCounter(&perfCnt2); }
	perfCnt1 = perfCnt2;

	// Copy the virtual screen to the window
	memset(&bi,0,sizeof(BITMAPINFOHEADER));
 	bi.bih.biSize = sizeof(BITMAPINFOHEADER);
 	bi.bih.biWidth = width;
 	bi.bih.biHeight = -visibleLines;
 	bi.bih.biPlanes = 1;
 	bi.bih.biBitCount = 16;
  	bi.bih.biCompression = BI_BITFIELDS;

    *((int*)&bi.mask[0]) = 0xF800;	// RRRRR000 00000000
    *((int*)&bi.mask[1]) = 0x07E0;	// 00000GGG GGG00000
    *((int*)&bi.mask[2]) = 0x001F;	// 00000000 000BBBBB

 	GetClientRect(hwnd,&rect);


 	StretchDIBits(dc,0,0,rect.right,rect.bottom-20,
 	               0,0,width,visibleLines,vscreen,(BITMAPINFO*)&bi,DIB_RGB_COLORS,SRCCOPY);

	zombie(TELL,GPU_FRAME_COMPLETED,0,0);
}


void reset_screen()
{
	static char buffer[80];
	yscroll = gpu.reg5;

/*	wsprintf(buffer,"Nametable[0..3] = 0x%08x\n",*((unsigned int*)&VRAM[NAME_TABLE_OFFSET]));
	WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/
}



BOOL CALLBACK DlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	static char stemp[64];

	return FALSE;
}
